@@ -7,10 +7,10 @@
 
 /*
  * DOCS: - "lo@" prefix for fstype is unsupported
- *	 - encyption= mount option for loop device is unssuported
  */
 
 #include <blkid.h>
+#include <stdbool.h>
 
 #include "mountP.h"
 #include "loopdev.h"
@@ -35,23 +35,23 @@ int mnt_context_is_loopdev(struct libmnt_context *cxt)
 
 	if (cxt->user_mountflags & (MNT_MS_LOOP |
 				    MNT_MS_OFFSET |
-				    MNT_MS_SIZELIMIT |
-				    MNT_MS_ENCRYPTION)) {
+				    MNT_MS_SIZELIMIT)) {
 
-		DBG(CXT, mnt_debug_h(cxt, "loopdev specific options detected"));
+		DBG(LOOP, ul_debugobj(cxt, "loopdev specific options detected"));
 		return 1;
 	}
 
-	if (cxt->mountflags & (MS_BIND | MS_MOVE | MS_PROPAGATION))
+	if ((cxt->mountflags & (MS_BIND | MS_MOVE))
+	    || mnt_context_propagation_only(cxt))
 		return 0;
 
 	/* Automatically create a loop device from a regular file if a
 	 * filesystem is not specified or the filesystem is known for libblkid
 	 * (these filesystems work with block devices only). The file size
-	 * should be at least 1KiB otherwise we will create empty loopdev where
-	 * is no mountable filesystem...
+	 * should be at least 1KiB, otherwise we will create an empty loopdev with
+	 * no mountable filesystem...
 	 *
-	 * Note that there is not a restriction (on kernel side) that prevents regular
+	 * Note that there is no restriction (on kernel side) that would prevent a regular
 	 * file as a mount(2) source argument. A filesystem that is able to mount
 	 * regular files could be implemented.
 	 */
@@ -63,7 +63,7 @@ int mnt_context_is_loopdev(struct libmnt_context *cxt)
 
 		if (stat(src, &st) == 0 && S_ISREG(st.st_mode) &&
 		    st.st_size > 1024) {
-			DBG(CXT, mnt_debug_h(cxt, "automatically enabling loop= option"));
+			DBG(LOOP, ul_debugobj(cxt, "automatically enabling loop= option"));
 			cxt->user_mountflags |= MNT_MS_LOOP;
 			mnt_optstr_append_option(&cxt->fs->user_optstr, "loop", NULL);
 			return 1;
@@ -74,10 +74,11 @@ int mnt_context_is_loopdev(struct libmnt_context *cxt)
 }
 
 
-/* Check, if there already exists a mounted loop device on the mountpoint node
+/* Check if there already exists a mounted loop device on the mountpoint node
  * with the same parameters.
  */
-static int is_mounted_same_loopfile(struct libmnt_context *cxt,
+static int __attribute__((nonnull))
+is_mounted_same_loopfile(struct libmnt_context *cxt,
 				    const char *target,
 				    const char *backing_file,
 				    uint64_t offset)
@@ -86,63 +87,63 @@ static int is_mounted_same_loopfile(struct libmnt_context *cxt,
 	struct libmnt_iter itr;
 	struct libmnt_fs *fs;
 	struct libmnt_cache *cache;
+	const char *bf;
+	int rc = 0;
 
 	assert(cxt);
 	assert(cxt->fs);
 	assert((cxt->flags & MNT_FL_MOUNTFLAGS_MERGED));
 
-	if (!target || !backing_file || mnt_context_get_mtab(cxt, &tb))
+	if (mnt_context_get_mtab(cxt, &tb))
 		return 0;
 
-	DBG(CXT, mnt_debug_h(cxt, "checking if %s mounted on %s",
+	DBG(LOOP, ul_debugobj(cxt, "checking if %s mounted on %s",
 				backing_file, target));
 
 	cache = mnt_context_get_cache(cxt);
 	mnt_reset_iter(&itr, MNT_ITER_BACKWARD);
 
-	/* Search for mountpoint node in mtab, procceed if any of these has the
+	bf = cache ? mnt_resolve_path(backing_file, cache) : backing_file;
+
+	/* Search for a mountpoint node in mtab, proceed if any of these have the
 	 * loop option set or the device is a loop device
 	 */
-	while (mnt_table_next_fs(tb, &itr, &fs) == 0) {
+	while (rc == 0 && mnt_table_next_fs(tb, &itr, &fs) == 0) {
 		const char *src = mnt_fs_get_source(fs);
 		const char *opts = mnt_fs_get_user_options(fs);
 		char *val;
 		size_t len;
-		int res = 0;
 
 		if (!src || !mnt_fs_match_target(fs, target, cache))
 			continue;
 
+		rc = 0;
+
 		if (strncmp(src, "/dev/loop", 9) == 0) {
-			res = loopdev_is_used((char *) src, backing_file,
-					offset, LOOPDEV_FL_OFFSET);
+			rc = loopdev_is_used((char *) src, bf, offset, 0, LOOPDEV_FL_OFFSET);
 
 		} else if (opts && (cxt->user_mountflags & MNT_MS_LOOP) &&
 		    mnt_optstr_get_option(opts, "loop", &val, &len) == 0 && val) {
 
 			val = strndup(val, len);
-			res = loopdev_is_used((char *) val, backing_file,
-					offset, LOOPDEV_FL_OFFSET);
+			rc = loopdev_is_used((char *) val, bf, offset, 0, LOOPDEV_FL_OFFSET);
 			free(val);
 		}
-
-		if (res) {
-			DBG(CXT, mnt_debug_h(cxt, "%s already mounted", backing_file));
-			return 1;
-		}
 	}
-
-	return 0;
+	if (rc)
+		DBG(LOOP, ul_debugobj(cxt, "%s already mounted", backing_file));
+	return rc;
 }
 
 int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 {
 	const char *backing_file, *optstr, *loopdev = NULL;
-	char *val = NULL, *enc = NULL, *pwd = NULL;
+	char *val = NULL, *loopval = NULL;
 	size_t len;
 	struct loopdev_cxt lc;
 	int rc = 0, lo_flags = 0;
 	uint64_t offset = 0, sizelimit = 0;
+	bool reuse = FALSE;
 
 	assert(cxt);
 	assert(cxt->fs);
@@ -152,19 +153,13 @@ int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 	if (!backing_file)
 		return -EINVAL;
 
-	DBG(CXT, mnt_debug_h(cxt, "trying to setup loopdev for %s", backing_file));
+	DBG(LOOP, ul_debugobj(cxt, "trying to setup device for %s", backing_file));
 
 	if (cxt->mountflags & MS_RDONLY) {
-		DBG(CXT, mnt_debug_h(cxt, "enabling READ-ONLY flag"));
+		DBG(LOOP, ul_debugobj(cxt, "enabling READ-ONLY flag"));
 		lo_flags |= LO_FLAGS_READ_ONLY;
 	}
 
-	rc = loopcxt_init(&lc, 0);
-	if (rc)
-		return rc;
-
-	ON_DBG(CXT, loopcxt_enable_debug(&lc, 1));
-
 	optstr = mnt_fs_get_user_options(cxt->fs);
 
 	/*
@@ -172,13 +167,8 @@ int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 	 */
 	if (rc == 0 && (cxt->user_mountflags & MNT_MS_LOOP) &&
 	    mnt_optstr_get_option(optstr, "loop", &val, &len) == 0 && val) {
-
-		val = strndup(val, len);
-		rc = val ? loopcxt_set_device(&lc, val) : -ENOMEM;
-		free(val);
-
-		if (rc == 0)
-			loopdev = loopcxt_get_device(&lc);
+		loopval = strndup(val, len);
+		rc = loopval ? 0 : -ENOMEM;
 	}
 
 	/*
@@ -188,7 +178,7 @@ int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 	    mnt_optstr_get_option(optstr, "offset", &val, &len) == 0) {
 		rc = mnt_parse_offset(val, len, &offset);
 		if (rc) {
-			DBG(CXT, mnt_debug_h(cxt, "failed to parse offset="));
+			DBG(LOOP, ul_debugobj(cxt, "failed to parse offset="));
 			rc = -MNT_ERR_MOUNTOPT;
 		}
 	}
@@ -200,7 +190,7 @@ int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 	    mnt_optstr_get_option(optstr, "sizelimit", &val, &len) == 0) {
 		rc = mnt_parse_offset(val, len, &sizelimit);
 		if (rc) {
-			DBG(CXT, mnt_debug_h(cxt, "failed to parse sizelimit="));
+			DBG(LOOP, ul_debugobj(cxt, "failed to parse sizelimit="));
 			rc = -MNT_ERR_MOUNTOPT;
 		}
 	}
@@ -210,13 +200,8 @@ int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 	 */
 	if (rc == 0 && (cxt->user_mountflags & MNT_MS_ENCRYPTION) &&
 	    mnt_optstr_get_option(optstr, "encryption", &val, &len) == 0) {
-		enc = strndup(val, len);
-		if (val && !enc)
-			rc = -ENOMEM;
-		if (enc && cxt->pwd_get_cb) {
-			DBG(CXT, mnt_debug_h(cxt, "asking for pass"));
-			pwd = cxt->pwd_get_cb(cxt);
-		}
+		DBG(LOOP, ul_debugobj(cxt, "encryption no longer supported"));
+		rc = -MNT_ERR_MOUNTOPT;
 	}
 
 	if (rc == 0 && is_mounted_same_loopfile(cxt,
@@ -224,6 +209,81 @@ int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 				backing_file, offset))
 		rc = -EBUSY;
 
+	if (rc)
+		goto done_no_deinit;
+
+	/* It is possible to mount the same file more times. If we set more
+	 * than one loop device referring to the same file, kernel has no
+	 * mechanism to detect it. To prevent data corruption, the same loop
+	 * device has to be recycled.
+	*/
+	if (backing_file) {
+		rc = loopcxt_init(&lc, 0);
+		if (rc)
+			goto done_no_deinit;
+
+		rc = loopcxt_find_overlap(&lc, backing_file, offset, sizelimit);
+		switch (rc) {
+		case 0: /* not found */
+			DBG(LOOP, ul_debugobj(cxt, "not found overlapping loopdev"));
+			loopcxt_deinit(&lc);
+			break;
+
+		case 1:	/* overlap */
+			DBG(LOOP, ul_debugobj(cxt, "overlapping %s detected",
+						loopcxt_get_device(&lc)));
+			rc = -MNT_ERR_LOOPOVERLAP;
+			goto done;
+
+		case 2: /* overlap -- full size and offset match (reuse) */
+		{
+			uint32_t lc_encrypt_type;
+
+			DBG(LOOP, ul_debugobj(cxt, "re-using existing loop device %s",
+				loopcxt_get_device(&lc)));
+
+			/* Once a loop is initialized RO, there is no
+			 * way to change its parameters. */
+			if (loopcxt_is_readonly(&lc)
+			    && !(lo_flags & LO_FLAGS_READ_ONLY)) {
+				DBG(LOOP, ul_debugobj(cxt, "%s is read-only",
+						loopcxt_get_device(&lc)));
+				rc = -EROFS;
+				goto done;
+			}
+
+			/* This is no more supported, but check to be safe. */
+			if (loopcxt_get_encrypt_type(&lc, &lc_encrypt_type) == 0
+			    && lc_encrypt_type != LO_CRYPT_NONE) {
+				DBG(LOOP, ul_debugobj(cxt, "encryption no longer supported for device %s",
+					loopcxt_get_device(&lc)));
+				rc = -MNT_ERR_LOOPOVERLAP;
+				goto done;
+			}
+			rc = 0;
+			/* loop= used with argument. Conflict will occur. */
+			if (loopval) {
+				rc = -MNT_ERR_LOOPOVERLAP;
+				goto done;
+			} else {
+				reuse = TRUE;
+				goto success;
+			}
+		}
+		default: /* error */
+			goto done;
+		}
+	}
+
+	DBG(LOOP, ul_debugobj(cxt, "not found; create a new loop device"));
+	rc = loopcxt_init(&lc, 0);
+	if (rc)
+		goto done_no_deinit;
+	if (loopval) {
+		rc = loopcxt_set_device(&lc, loopval);
+		if (rc == 0)
+			loopdev = loopcxt_get_device(&lc);
+	}
 	if (rc)
 		goto done;
 
@@ -231,8 +291,8 @@ int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 	 * because kernel provides the name in /sys.
 	 */
 	if (get_linux_version() >= KERNEL_VERSION(2, 6, 37) ||
-	    !cxt->mtab_writable) {
-		DBG(CXT, mnt_debug_h(cxt, "enabling AUTOCLEAR flag"));
+	    !mnt_context_mtab_writable(cxt)) {
+		DBG(LOOP, ul_debugobj(cxt, "enabling AUTOCLEAR flag"));
 		lo_flags |= LO_FLAGS_AUTOCLEAR;
 	}
 
@@ -242,7 +302,7 @@ int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 			rc = loopcxt_find_unused(&lc);
 			if (rc)
 				goto done;
-			DBG(CXT, mnt_debug_h(cxt, "trying to use %s",
+			DBG(LOOP, ul_debugobj(cxt, "trying to use %s",
 						loopcxt_get_device(&lc)));
 		}
 
@@ -255,12 +315,10 @@ int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 			rc = loopcxt_set_offset(&lc, offset);
 		if (!rc && sizelimit)
 			rc = loopcxt_set_sizelimit(&lc, sizelimit);
-		if (!rc && enc && pwd)
-			loopcxt_set_encryption(&lc, enc, pwd);
 		if (!rc)
 			loopcxt_set_flags(&lc, lo_flags);
 		if (rc) {
-			DBG(CXT, mnt_debug_h(cxt, "failed to set loopdev attributes"));
+			DBG(LOOP, ul_debugobj(cxt, "failed to set loop attributes"));
 			goto done;
 		}
 
@@ -270,13 +328,14 @@ int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 			break;		/* success */
 
 		if (loopdev || rc != -EBUSY) {
-			DBG(CXT, mnt_debug_h(cxt, "failed to setup device"));
+			DBG(LOOP, ul_debugobj(cxt, "failed to setup device"));
 			rc = -MNT_ERR_LOOPDEV;
 			goto done;
 		}
-		DBG(CXT, mnt_debug_h(cxt, "loopdev stolen...trying again"));
+		DBG(LOOP, ul_debugobj(cxt, "device stolen...trying again"));
 	} while (1);
 
+success:
 	if (!rc)
 		rc = mnt_fs_set_source(cxt->fs, loopcxt_get_device(&lc));
 
@@ -284,12 +343,13 @@ int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 		/* success */
 		cxt->flags |= MNT_FL_LOOPDEV_READY;
 
-		if ((cxt->user_mountflags & MNT_MS_LOOP) &&
-		    loopcxt_is_autoclear(&lc)) {
+		if (reuse || ( (cxt->user_mountflags & MNT_MS_LOOP) &&
+		    loopcxt_is_autoclear(&lc))) {
 			/*
-			 * autoclear flag accepted by kernel, don't store
+			 * autoclear flag accepted by the kernel, don't store
 			 * the "loop=" option to mtab.
 			 */
+			DBG(LOOP, ul_debugobj(cxt, "removing unnecessary loop= from mtab"));
 			cxt->user_mountflags &= ~MNT_MS_LOOP;
 			mnt_optstr_remove_option(&cxt->fs->user_optstr, "loop");
 		}
@@ -303,18 +363,19 @@ int mnt_context_setup_loopdev(struct libmnt_context *cxt)
 			mnt_context_set_mflags(cxt, cxt->mountflags | MS_RDONLY);
 
 		/* we have to keep the device open until mount(1),
-		 * otherwise it will auto-cleared by kernel
+		 * otherwise it will be auto-cleared by kernel
 		 */
 		cxt->loopdev_fd = loopcxt_get_fd(&lc);
-		loopcxt_set_fd(&lc, -1, 0);
+		if (cxt->loopdev_fd < 0) {
+			DBG(LOOP, ul_debugobj(cxt, "failed to get loopdev FD"));
+			rc = -errno;
+		} else
+			loopcxt_set_fd(&lc, -1, 0);
 	}
 done:
-	free(enc);
-	if (pwd && cxt->pwd_release_cb) {
-		DBG(CXT, mnt_debug_h(cxt, "release pass"));
-		cxt->pwd_release_cb(cxt, pwd);
-	}
 	loopcxt_deinit(&lc);
+done_no_deinit:
+	free(loopval);
 	return rc;
 }
 
@@ -340,7 +401,7 @@ int mnt_context_delete_loopdev(struct libmnt_context *cxt)
 	cxt->flags &= ~MNT_FL_LOOPDEV_READY;
 	cxt->loopdev_fd = -1;
 
-	DBG(CXT, mnt_debug_h(cxt, "loopdev deleted [rc=%d]", rc));
+	DBG(LOOP, ul_debugobj(cxt, "deleted [rc=%d]", rc));
 	return rc;
 }
 
@@ -363,7 +424,7 @@ int mnt_context_clear_loopdev(struct libmnt_context *cxt)
 		/*
 		 * mount(2) success, close the device
 		 */
-		DBG(CXT, mnt_debug_h(cxt, "closing loopdev FD"));
+		DBG(LOOP, ul_debugobj(cxt, "closing FD"));
 		close(cxt->loopdev_fd);
 	}
 	cxt->loopdev_fd = -1;
